from typing import Set

from Bio.Seq import Seq
from Bio.SeqFeature import FeatureLocation, CompoundLocation


class Transcript(object):
    def __init__(self, transcript_id, chromosome = None, start = 0, end = 0, strand = None, features = list(), **gff_attributes):
        self.transcript_id = transcript_id
        self.chromosome = chromosome
        self.attributes = gff_attributes
        self.features = features
        
        if strand in ("-1", -1, "-"):
            strand = -1
        elif strand in ("+1", "1", 1, "+"):
            strand = 1
        
        self.location = FeatureLocation(start = start, end = end, strand = strand, ref = transcript_id)
        
        # cached properties
        self._polypeptide = None
        self._exons = None
        self._five_prime_utr = None
        self._three_prime_utr = None
        
        self._nucleic_coding_sequence = None
        self._protein_sequence = None
    
    def __repr__(self):
        return u"%s" % self.transcript_id
    
    def get(self, attribute):
        return self.__dict__.get(attribute, None)
    
    def get_all_ids(self) -> Set[str]:
        all_ids = set()  # all possible IDs for the given gene
        
        transcript_name = self.attributes.get('Name', None)
        if transcript_name:
            all_ids.add(transcript_name)
        
        transcript_id = self.attributes.get('ID', None)
        if transcript_id:
            all_ids.add(transcript_id)
        
        ancestor_identifier = self.attributes.get('ancestorIdentifier', None)
        if ancestor_identifier:
            all_ids.add(ancestor_identifier)
        
        if self.protein_id:
            all_ids.add(self.protein_id)
        
        return all_ids
    
    @property
    def protein_id(self):
        try:
            polypeptide_feature = next(filter(lambda f: f['type'] == 'polypeptide', self.features))
            return polypeptide_feature['attributes'].get('Name', polypeptide_feature['attributes'].get('ID', None))
        except StopIteration:
            return None
    
    @property
    def protein_sequence(self) -> Seq:
        """
        Return the protein sequence for this transcript using CDS annotations
        """
        if not self._protein_sequence:
            nucleic_sequence = self.nucleic_coding_sequence
            
            if not nucleic_sequence:
                return Seq("")
            
            # make sure the sequence length is a multiple of 3
            end_position = len(nucleic_sequence) - len(nucleic_sequence) % 3
            if isinstance(nucleic_sequence, str):
                nucleic_sequence = Seq(nucleic_sequence[:end_position])
            else:
                nucleic_sequence = nucleic_sequence[:end_position]
            
            self._protein_sequence = nucleic_sequence.translate()
        
        return self._protein_sequence
    
    def nucleic_sequence(self, feature_type):
        """
        :type feature_type: str
        :rtype: Bio.Seq.Seq
        """
        location = self._get_features(feature_type)
        if location:
            return location.extract(self.chromosome.nucleic_sequence)
        return None
    
    @property
    def nucleic_coding_sequence(self) -> Seq:
        if not self._nucleic_coding_sequence:
            if self.cds:
                self._nucleic_coding_sequence = self.cds.extract(self.chromosome.nucleic_sequence)
                if self.location.strand == -1:
                    self._nucleic_coding_sequence = Seq(self._nucleic_coding_sequence).reverse_complement()
        
        return self._nucleic_coding_sequence
    
    @property
    def cds(self) -> CompoundLocation:
        if not hasattr(self, "_cds"):
            cds_list = sorted(
                map(
                    lambda exon: FeatureLocation(
                        start = max(exon.start, self.polypeptide.start),
                        end = min(exon.end, self.polypeptide.end)
                    ),
                    filter(
                        lambda exon: exon.start <= self.polypeptide.end and exon.end >= self.polypeptide.start,
                        self.exons.parts
                    ),
                ),
                key = lambda exon: exon.start
            )
            
            if len(cds_list) == 1:
                self._cds = cds_list[0]
            
            if len(cds_list) > 1:
                self._cds = CompoundLocation(cds_list)
        
        return self._cds
    
    @property
    def exons(self):
        if self._exons is None:
            self._exons = self._get_features("exon")
        return self._exons
    
    @property
    def polypeptide(self):
        if self._polypeptide is None:
            self._polypeptide = self._get_features("polypeptide")
        
        return self._polypeptide
    
    @property
    def five_prime_utr(self):
        if self._five_prime_utr is None:
            utr_list = sorted(
                map(
                    lambda exon: FeatureLocation(
                        start = exon.start,
                        end = min(exon.end, self.polypeptide.start)
                    ),
                    filter(
                        lambda exon: exon.start <= self.polypeptide.start,
                        self.exons.parts
                    )
                ),
                key = lambda exon: exon.start
            )
            
            if len(utr_list) == 1:
                self._five_prime_utr = utr_list[0]
            
            elif len(utr_list) > 1:
                self._five_prime_utr = CompoundLocation(utr_list)
        
        return self._five_prime_utr
    
    @property
    def three_prime_utr(self):
        return self._get_features("three_prime_utr")
    
    def _get_features(self, feature_type:str):
        locations = []
        features = sorted(
            filter(lambda f: f['type'].lower() == feature_type, self.features),
            key = lambda f: f.get('start', 0)
        )
        for feature in features:
            locations.append(FeatureLocation(feature['start'], feature['end']))
        
        if len(locations) == 1:
            return locations[0]
        if len(locations) > 1:
            return CompoundLocation(locations)
        if len(features) > 0:
            return features[0]
        return None



import logging
import sys
import os


class Log:
    def __init__(self, logger_name=None, log_level='INFO', log_dir=None, logger:logging.Logger=None, **kwargs):
        self.logger:logging.Logger = None
        self.logger_name = logger_name
        self.log_level = log_level
        self.log_dir = log_dir
        
        if logger and isinstance(logger, logging.Logger):
            self.logger = logger
        elif self.logger_name:
            # retrieve existing logger if already set
            if self.logger_name in logging.Logger.manager.loggerDict.keys():
                self.logger = logging.getLogger(self.logger_name)
            
            # set filehandler logger
            elif self.log_dir:
                if not os.path.exists(log_dir):
                    os.makedirs(log_dir, exist_ok=True)
                log_filepath = os.path.join(self.log_dir, self.logger_name + '.log')
                self.logger = self.get_file_logger(filepath=log_filepath)
            
            # set stream stdout logger
            else:
                self.logger = self.get_stream_logger()

    def get_file_logger(self, filepath) -> logging.Logger:
        handler_formatter = logging.Formatter(fmt='%(message)s')

        handler = logging.FileHandler(filename=filepath, mode="w", encoding='utf8')
        handler.setFormatter(handler_formatter)
        handler.setLevel(self.log_level)

        logger = logging.getLogger(self.logger_name)
        logger.setLevel(self.log_level)
        logger.addHandler(handler)

        return logger

    def get_stream_logger(self, fmt='%(asctime)s  %(message)s') -> logging.Logger:
        handler_formatter = logging.Formatter(fmt=fmt, datefmt='[%Y/%m/%d %I:%M:%S]')

        handler = logging.StreamHandler(stream=sys.stdout)
        handler.setFormatter(handler_formatter)
        handler.setLevel(self.log_level)

        logger = logging.getLogger(self.logger_name)
        logger.setLevel(self.log_level)
        logger.addHandler(handler)

        return logger

    def log(self, *msg, **kwargs):
        level = kwargs.get('level', 'INFO')
        
        # convert level to code if nedeed
        if isinstance(level, str):
            level = getattr(logging, level)

        if self.logger:
            self.logger.log(level, " ".join(msg))
        else:
            print(*msg)



from urllib import request, parse
import os
import time
import json
import subprocess

from bioinfo_tools.utils.log import Log
from bioinfo_tools.utils.sge import SgeJob


def fetch_taxonomy(species_name):
    url = "https://eutils.ncbi.nlm.nih.gov/entrez/eutils/esearch.fcgi"
    params = parse.urlencode({
        'db': 'taxonomy',
        'term': species_name,
        'retmode': 'json'
    })
    uri = url + '?' + params
    
    with request.urlopen(uri) as r:
        ncbi_response = r.read()
    
    json_response = json.loads(ncbi_response.decode())
    
    ncbi_id = None
    ncbi_xml = None
    
    if int(json_response.get("esearchresult", {}).get("count", 0)) > 0:
        ncbi_id = json_response.get("esearchresult", {}).get("idlist", [])[0]
        if ncbi_id:
            url = "https://eutils.ncbi.nlm.nih.gov/entrez/eutils/efetch.fcgi?db=taxonomy&id=" + ncbi_id
            with request.urlopen(url) as r:
                ncbi_response = r.read()
            ncbi_xml = ncbi_response.decode()
    
    return ncbi_id, ncbi_xml


def makeblastdb(fasta_filepath, dbtype = 'prot'):
    cmd = [
        "makeblastdb",
        "-in", fasta_filepath,
        "-dbtype", dbtype
    ]
    subprocess.check_call(" ".join(cmd), shell = True)


class BlastCommand(Log):
    def __init__(self, *args, **kwargs):
        super().__init__(**kwargs)
        
        self.blast_prog = None
        self.sge_job = None
        
        for arg_name in ['query', 'db', 'out']:
            if arg_name not in kwargs:
                raise Exception("'%s' argument is required" % arg_name)

        self.args = kwargs
        if not os.path.exists(self.args['query']):
            raise Exception("query file not found: " + self.args['query'])

        if isinstance(self.args['db'], list):
            self.args['db'] = "'" + " ".join(self.args['db']) + "'"
    
    def run(self, job_name = None, use_sge = False, async = False):
        """
        run blast command, either locally or through qsub
        """
        blast_commandline = self._create_commandline()

        if use_sge:
            if not job_name:
                job_name = os.path.basename(self.args['out']).split('.')[0]
            self.sge_job = SgeJob()
            self.sge_job.submit(blast_commandline, job_name = '%s_%s' % (self.blast_prog, job_name), sync = not async)
            return self.sge_job.get_job_id()
        else:
            if async:
                # TODO: make this option available locally
                self.log("Can't run in async mode locally. Will run synchronously.")
            
            try:
                return subprocess.check_output(blast_commandline, shell = True)
            except subprocess.CalledProcessError as e:
                raise Exception(e.output)
    
    def run_async(self, job_name = None, use_sge = False):
        return self.run(job_name = job_name, use_sge = use_sge, async = True)
    
    def wait_for_output(self, max_checks = 10):
        """
        wait for output file to be writen on disk
        """
        while not os.path.exists(self.args['out']) and max_checks > 0:
            time.sleep(10)
            max_checks -= 1
    
        if not os.path.exists(self.args['out']):
            raise Exception("blast output does not exist: " + self.args['out'])

    def _create_commandline(self):
        if not self.blast_prog:
            raise Exception("blast_prog must be set")
    
        blast_params = []
        for key, value in self.args.items():
            if not key.startswith('-'):
                key = '-' + key
            blast_params.append(key + '=' + str(value))
        return self.blast_prog + " " + " ".join(blast_params)


class Blastp(BlastCommand):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.blast_prog = 'blastp'


class Blastx(BlastCommand):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.blast_prog = 'blastx'


class Blastn(BlastCommand):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.blast_prog = 'blastn'



import copy
import re

from bioinfo_tools.utils.log import Log


class OboParser(Log):
    DEFAULT_KEPT_FIELDS = ['id', 'name', 'namespace', 'def', 'synonym', 'is_a', 'alt_id']

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.terms = {}

    def read(self, file_handler, kept_fields = None, prefix_with = None):
        """
        Parses a OBO file v1.2 format.
        return all Term entries
        """
        self.terms = {}
        kept_fields = kept_fields is not None and kept_fields or self.DEFAULT_KEPT_FIELDS
    
        current_term = None
        for line in file_handler:
            line = line.strip()
            if not line:
                continue  # Skip empty

            if line == "[Term]":
                if current_term:
                    self.terms[current_term['id']] = current_term
                    for alt_id in current_term.get('alt_id', []):
                        self.terms[alt_id] = copy.deepcopy(current_term)
                        self.terms[alt_id]["id"] = alt_id
                current_term = {}

            elif line == "[Typedef]":  # Skip [Typedef] sections
                current_term = None

            elif current_term is not None:  # Only process if we're inside a [Term] environment
                key, value = line.split(": ", 1)
                
                if key not in kept_fields:
                    continue

                value = value.strip()

                m = re.search('"(.*)"', value)  # only keep text wrapped in double quotes (if any)
                if m:
                    value = m.group(1)

                if key == 'is_a':
                    value = value.split(' ! ')[0]

                if key not in current_term:
                    current_term[key] = value
                    if key in ['is_a', 'synonym']:  # force storing of these keys as list
                        current_term[key] = [value]
                else:
                    if not isinstance(current_term[key], list):
                        current_term[key] = [current_term[key]]
                    current_term[key].append(value)

        # Add last term
        if current_term is not None:
            self.terms[current_term['id']] = current_term

        if prefix_with:
            for term_id, term in self.terms.items():
                for key in term.keys():
                    if not key.startswith(prefix_with):
                        self.terms[term_id][prefix_with + key] = term.pop(key, None)

        return self.terms

    def get_parents(self, term, distance_from_child = 1):
        parents = []

        for term_id in term.get('is_a', []):
            parent = copy.deepcopy(self.terms[term_id])
            parent['distance'] = distance_from_child
            parents.append(parent)
            parents.extend(self.get_parents(parent, distance_from_child + 1))

        return parents



#!/usr/bin/env python
from typing import List, Dict

from Bio.SeqFeature import FeatureLocation
from Bio.SeqRecord import SeqRecord

from bioinfo_tools.genomic_features.transcript import Transcript


class Gene(object):
    def __init__(self, gene_id, chromosome = None, start = 0, end = 0, strand = None, assembly_name = None, **gff_attributes):
        self.gene_id = gene_id
        self.chromosome = chromosome
        self.attributes = gff_attributes
        self.transcripts = list()
        self.assembly_name = assembly_name
        
        if strand in ("-1", -1, "-"):
            self.strand = -1
        elif strand in ("+1", "1", 1, "+"):
            self.strand = 1
        
        self.location = FeatureLocation(start = start, end = end, strand = self.strand, ref = gene_id)
    
    def __repr__(self):
        return u"%s" % self.gene_id
    
    def __getitem__(self, item):
        return getattr(self, item)
    
    def __setitem__(self, key, value):
        setattr(self, key, value)
    
    def __eq__(self, other):
        if isinstance(other, str):
            return self.gene_id == other
        elif isinstance(other, Gene):
            return self.gene_id == other.gene_id
        return self == other
    
    def get(self, *args):
        return self.__dict__.get(*args)
    
    def add_transcript(self, mRNA_feature):
        transcript_id = None
        
        if 'transcript_id' in mRNA_feature:
            transcript_id = mRNA_feature.pop('transcript_id')
        elif 'attributes' in mRNA_feature:
            if 'transcript_id' in mRNA_feature['attributes']:
                transcript_id = mRNA_feature['attributes'].pop('transcript_id')
            if 'Name' in mRNA_feature['attributes']:
                transcript_id = mRNA_feature['attributes']['Name']
            elif 'ID' in mRNA_feature['attributes']:
                transcript_id = mRNA_feature['attributes']['ID']

        # remove all potential duplicated keys
        for key_name in ('chromosome', 'start', 'end', 'strand'):
            mRNA_feature.get('attributes', {}).pop(key_name, None)

        transcript = Transcript(
            transcript_id = transcript_id,
            chromosome = self.chromosome,
            start = mRNA_feature.get('start', 0),
            end = mRNA_feature.get('end', 0),
            strand = mRNA_feature.get('strand', None),
            features = mRNA_feature.get('features', []),
            **mRNA_feature.get('attributes', {})
        )
        self.transcripts.append(transcript)
    
    def as_fasta(self, **kwargs):
        record = SeqRecord(self.extract_sequence(**kwargs), id = self.gene_id)
        return record.format("fasta")
    
    def extract_sequence(self, upstream = 0, downstream = 0):
        location = FeatureLocation(self.location.start - upstream, self.location.end + downstream, self.strand)
        return location.extract(self.chromosome.nucleic_sequence)

    def get_all_ids(self) -> List[str]:
        """
        return all possible IDs for that gene
        :rtype: set
        """
        if not hasattr(self, '_all_ids'):
            all_ids = set()  # all possible IDs for the given gene
            
            gene_name = self.attributes.get('Name', None)
            if gene_name:
                all_ids.add(gene_name)
            
            gene_id = self.attributes.get('ID', None)
            if gene_id:
                all_ids.add(gene_id)
    
            ancestor_identifier = self.attributes.get('ancestorIdentifier', None)
            if ancestor_identifier:
                all_ids.add(ancestor_identifier)
    
            for transcript in self.transcripts:
                all_ids.update(transcript.get_all_ids())
        
            self._all_ids = list(all_ids)
        
        return self._all_ids



from typing import Tuple, Iterable

import re
from Bio import SeqIO


class FastaParser(object):
    def __init__(self):
        pass
    
    def read(self, fasta_file: str, id_separator = "\||\:") -> Iterable[Tuple[str,str]]:
        with open(fasta_file) as fh:
            for record in SeqIO.parse(fh, "fasta"):
                seqid = re.split(id_separator, record.id, 1)[1]
                seq = str(record.seq)
                yield (seqid, seq)



from typing import Dict, List, Set

from Bio.SeqFeature import FeatureLocation

from bioinfo_tools.genomic_features.gene import Gene


class Chromosome(object):
    def __init__(self, chromosome_id, assembly_name = None):
        self.chromosome_id = chromosome_id
        self.assembly_name = assembly_name
        self.length = 0
        self._genes = {}
        self._index_by_position = dict()
    
    def __repr__(self):
        return u"%s (%s genes)" % (self.chromosome_id, len(self.genes))
    
    def add_gene(self, gff_feature:Dict):
        gene_id = None
        
        if 'gene_id' in gff_feature:
            gene_id = gff_feature.pop('gene_id')
        elif 'attributes' in gff_feature:
            if 'gene_id' in gff_feature['attributes']:
                gene_id = gff_feature['attributes'].pop('gene_id')
            elif 'Name' in gff_feature['attributes']:
                gene_id = gff_feature['attributes']['Name']
            elif 'ID' in gff_feature['attributes']:
                gene_id = gff_feature['attributes']['ID']
        
        if not gene_id:
            raise Exception("gene_id not found in given GFF feature")
        
        # remove all potential duplicated keys
        for key_name in ('chromosome', 'start', 'end', 'strand', 'assembly_name'):
            gff_feature.get('attributes', {}).pop(key_name, None)
        
        gene = Gene(
            gene_id = gene_id,
            chromosome = self,
            start = gff_feature.get('start', 0),
            end = gff_feature.get('end', 0),
            strand = gff_feature.get('strand', None),
            assembly_name = self.assembly_name,
            **gff_feature.get('attributes', {})
        )
        for transcript in gff_feature.get('mRNA', []):
            gene.add_transcript(transcript)
        
        self._genes[gene['gene_id']] = gene
        self._add_gene_to_index(gene)
        return gene
    
    @property
    def genes(self) -> List[Gene]:
        return list(self._genes.values())
    
    @property
    def sorted_genes(self) -> List[Gene]:
        if not hasattr(self, "_sorted_genes"):
            self._sorted_genes = sorted(self._genes.values(), key = lambda gene: gene.location.start)
        return self._sorted_genes
    
    def get_genes_at(self, position) -> List[Gene]:
        if not self._index_by_position:
            self.build_index()
        found_genes = list()
        for interval, gene in self._index_by_position.items():
            if position in interval and gene not in found_genes:
                found_genes.append(gene)
        return found_genes
    
    def get_gene(self, gene_id) -> Gene:
        return self._genes.get(gene_id, None)

    def build_index(self):
        self._index_by_position = dict()
        for gene in self.genes:
            self._add_gene_to_index(gene)
    
    def _add_gene_to_index(self, gene:Gene):
        if not hasattr(self, "_index_by_position"):
            self._index_by_position = dict()
        self._index_by_position[range(gene.location.start, gene.location.end + 1)] = gene

    def attach_nucleic_sequence(self, sequence):
        self.nucleic_sequence = sequence
        self.length = len(sequence)
    
    def extract_sequence(self, start, end):
        location = FeatureLocation(start, end)
        return location.extract(self.nucleic_sequence)



from setuptools import setup
import sys

if sys.version_info < (3,6):
    sys.exit('Sorry, Python < 3.6 is not supported')

setup(
    name='bioinfo_tools',
    version='0.2',
    url='https://github.com/sebriois/bioinfo_tools',
    author='Sebastien Briois, Guillaume Tiberi',
    author_email='sebriois@gmail.com',
    packages=[
        'bioinfo_tools',
        'bioinfo_tools.genomic_features',
        'bioinfo_tools.parsers',
        'bioinfo_tools.utils',
    ],
    keywords='bioinformatics',
    license='BSD',
    description='Python library that parses GFF, Fasta files into python classes',
    long_description=open('README.txt').read(),
    install_requires=[ "biopython" ],
    classifiers=[
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
    ],
)